"""Synchronous HTTP RPC-based message transports."""

from vumi.transports.vodacom_messaging.vodacom_messaging import (
    VodacomMessagingTransport, VodacomMessagingResponse)

__all__ = ['VodacomMessagingTransport', 'VodacomMessagingResponse']
